@@ -61,16 +61,17 @@ final class HttpRequestConduit extends AbstractStreamSinkConduit<StreamSinkCondu
     private final ClientRequest request;
 
     private static final int STATE_BODY = 0; // Message body, normal pass-through operation
-    private static final int STATE_START = 1; // No headers written yet
-    private static final int STATE_HDR_NAME = 2; // Header name indexed by charIndex
-    private static final int STATE_HDR_D = 3; // Header delimiter ':'
-    private static final int STATE_HDR_DS = 4; // Header delimiter ': '
-    private static final int STATE_HDR_VAL = 5; // Header value
-    private static final int STATE_HDR_EOL_CR = 6; // Header line CR
-    private static final int STATE_HDR_EOL_LF = 7; // Header line LF
-    private static final int STATE_HDR_FINAL_CR = 8; // Final CR
-    private static final int STATE_HDR_FINAL_LF = 9; // Final LF
-    private static final int STATE_BUF_FLUSH = 10; // flush the buffer and go to writing body
+    private static final int STATE_URL = 1; //Writing the URL
+    private static final int STATE_START = 2; // No headers written yet
+    private static final int STATE_HDR_NAME = 3; // Header name indexed by charIndex
+    private static final int STATE_HDR_D = 4; // Header delimiter ':'
+    private static final int STATE_HDR_DS = 5; // Header delimiter ': '
+    private static final int STATE_HDR_VAL = 6; // Header value
+    private static final int STATE_HDR_EOL_CR = 7; // Header line CR
+    private static final int STATE_HDR_EOL_LF = 8; // Header line LF
+    private static final int STATE_HDR_FINAL_CR = 9; // Final CR
+    private static final int STATE_HDR_FINAL_LF = 10; // Final LF
+    private static final int STATE_BUF_FLUSH = 11; // flush the buffer and go to writing body
 
     private static final int MASK_STATE         = 0x0000000F;
     private static final int FLAG_SHUTDOWN      = 0x00000010;
@@ -126,18 +127,34 @@ final class HttpRequestConduit extends AbstractStreamSinkConduit<StreamSinkCondu
                 }
                 case STATE_START: {
                     log.trace("Starting request");
-                    // we assume that our buffer has enough space for the initial request line plus one more CR+LF
-                    assert buffer.remaining() >= 50;
-                    request.getMethod().appendTo(buffer);
-                    buffer.put((byte) ' ');
-                    string = request.getPath();
-                    length = string.length();
-                    for (charIndex = 0; charIndex < length; charIndex ++) {
-                        buffer.put((byte) string.charAt(charIndex));
+                    int len = request.getMethod().length() + request.getPath().length() + request.getProtocol().length() + 4;
+
+                    // test that our buffer has enough space for the initial request line plus one more CR+LF
+                    if(len <= buffer.remaining()) {
+                        assert buffer.remaining() >= 50;
+                        request.getMethod().appendTo(buffer);
+                        buffer.put((byte) ' ');
+                        string = request.getPath();
+                        length = string.length();
+                        for (charIndex = 0; charIndex < length; charIndex++) {
+                            buffer.put((byte) string.charAt(charIndex));
+                        }
+                        buffer.put((byte) ' ');
+                        request.getProtocol().appendTo(buffer);
+                        buffer.put((byte) '\r').put((byte) '\n');
+                    } else {
+                        StringBuilder sb = new StringBuilder(len);
+                        sb.append(request.getMethod().toString());
+                        sb.append(" ");
+                        sb.append(request.getPath());
+                        sb.append(" ");
+                        sb.append(request.getProtocol());
+                        sb.append("\r\n");
+                        string = sb.toString();
+                        charIndex = 0;
+                        state = STATE_URL;
+                        break;
                     }
-                    buffer.put((byte) ' ');
-                    request.getProtocol().appendTo(buffer);
-                    buffer.put((byte) '\r').put((byte) '\n');
                     HeaderMap headers = request.getRequestHeaders();
                     nameIterator = headers.getHeaderNames().iterator();
                     if (! nameIterator.hasNext()) {
@@ -441,6 +458,48 @@ final class HttpRequestConduit extends AbstractStreamSinkConduit<StreamSinkCondu
                     pooledBuffer = null;
                     return STATE_BODY;
                 }
+                case STATE_URL: {
+                    for(int i = charIndex; i < string.length(); ++i) {
+                        if(!buffer.hasRemaining()) {
+                            buffer.flip();
+                            do {
+                                res = next.write(buffer);
+                                if (res == 0) {
+                                    log.trace("Continuation");
+                                    this.charIndex = i;
+                                    this.string = string;
+                                    this.state = STATE_URL;
+                                    return STATE_URL;
+                                }
+                            } while (buffer.hasRemaining());
+                            buffer.clear();
+                        }
+                        buffer.put((byte) string.charAt(i));
+                    }
+
+                    HeaderMap headers = request.getRequestHeaders();
+                    nameIterator = headers.getHeaderNames().iterator();
+                    state = STATE_HDR_NAME;
+                    if (! nameIterator.hasNext()) {
+                        log.trace("No request headers");
+                        buffer.put((byte) '\r').put((byte) '\n');
+                        buffer.flip();
+                        while (buffer.hasRemaining()) {
+                            res = next.write(buffer);
+                            if (res == 0) {
+                                log.trace("Continuation");
+                                return STATE_BUF_FLUSH;
+                            }
+                        }
+                        pooledBuffer.close();
+                        pooledBuffer = null;
+                        log.trace("Body");
+                        return STATE_BODY;
+                    }
+                    headerName = nameIterator.next();
+                    charIndex = 0;
+                    break;
+                }
                 default: {
                     throw new IllegalStateException();
                 }
@@ -612,7 +671,7 @@ final class HttpRequestConduit extends AbstractStreamSinkConduit<StreamSinkCondu
             }
             return;
         }
-        this.state = oldVal & ~MASK_STATE | FLAG_SHUTDOWN | STATE_BODY;
+        this.state = oldVal & ~MASK_STATE | FLAG_SHUTDOWN;
         throw new TruncatedResponseException();
     }
 
