@@ -21,6 +21,8 @@ package io.undertow.protocols.http2;
 import io.undertow.UndertowLogger;
 import io.undertow.UndertowMessages;
 import io.undertow.UndertowOptions;
+import io.undertow.connector.ByteBufferPool;
+import io.undertow.connector.PooledByteBuffer;
 import io.undertow.server.protocol.framed.AbstractFramedChannel;
 import io.undertow.server.protocol.framed.FrameHeaderData;
 import io.undertow.server.protocol.http2.Http2OpenListener;
@@ -34,26 +36,27 @@ import org.xnio.ChannelListener;
 import org.xnio.ChannelListeners;
 import org.xnio.IoUtils;
 import org.xnio.OptionMap;
-import io.undertow.connector.ByteBufferPool;
-import io.undertow.connector.PooledByteBuffer;
+import io.undertow.util.HttpString;
 import org.xnio.StreamConnection;
 import org.xnio.channels.StreamSinkChannel;
 import org.xnio.ssl.SslConnection;
 
-import javax.net.ssl.SSLSession;
 import java.io.IOException;
 import java.nio.ByteBuffer;
 import java.nio.channels.Channel;
 import java.nio.channels.ClosedChannelException;
+import java.security.SecureRandom;
 import java.util.ArrayList;
 import java.util.Collections;
 import java.util.HashMap;
 import java.util.List;
 import java.util.Map;
+import java.util.Random;
 import java.util.concurrent.ConcurrentHashMap;
+import javax.net.ssl.SSLSession;
 
 /**
- * SPDY channel.
+ * HTTP2 channel.
  *
  * @author Stuart Douglas
  */
@@ -61,6 +64,13 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
 
     public static final String CLEARTEXT_UPGRADE_STRING = "h2c";
 
+
+    public static final HttpString METHOD = new HttpString(":method");
+    public static final HttpString PATH = new HttpString(":path");
+    public static final HttpString SCHEME = new HttpString(":scheme");
+    public static final HttpString AUTHORITY = new HttpString(":authority");
+    public static final HttpString STATUS = new HttpString(":status");
+
     static final int FRAME_TYPE_DATA = 0x00;
     static final int FRAME_TYPE_HEADERS = 0x01;
     static final int FRAME_TYPE_PRIORITY = 0x02;
@@ -70,7 +80,7 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
     static final int FRAME_TYPE_PING = 0x06;
     static final int FRAME_TYPE_GOAWAY = 0x07;
     static final int FRAME_TYPE_WINDOW_UPDATE = 0x08;
-    static final int FRAME_TYPE_CONTINUATION = 0x09; //hopefully this goes away
+    static final int FRAME_TYPE_CONTINUATION = 0x09;
 
 
     public static final int ERROR_NO_ERROR = 0x00;
@@ -106,17 +116,17 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
 
     static final int DEFAULT_INITIAL_WINDOW_SIZE = 65535;
 
-    public static final byte[] PREFACE_BYTES = {
+    static final byte[] PREFACE_BYTES = {
             0x50, 0x52, 0x49, 0x20, 0x2a, 0x20, 0x48, 0x54,
             0x54, 0x50, 0x2f, 0x32, 0x2e, 0x30, 0x0d, 0x0a,
             0x0d, 0x0a, 0x53, 0x4d, 0x0d, 0x0a, 0x0d, 0x0a};
     public static final int DEFAULT_MAX_FRAME_SIZE = 16384;
     public static final int MAX_FRAME_SIZE = 16777215;
+    public static final int FLOW_CONTROL_MIN_WINDOW = 2;
 
 
     private Http2FrameHeaderParser frameParser;
-    private final Map<Integer, Http2StreamSourceChannel> incomingStreams = new ConcurrentHashMap<>();
-    private final Map<Integer, Http2StreamSinkChannel> outgoingStreams = new ConcurrentHashMap<>();
+    private final Map<Integer, StreamHolder> currentStreams = new ConcurrentHashMap<>();
     private final String protocol;
 
     //local
@@ -124,11 +134,9 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
     private boolean pushEnabled;
     private volatile int initialSendWindowSize = DEFAULT_INITIAL_WINDOW_SIZE;
     private volatile int initialReceiveWindowSize = DEFAULT_INITIAL_WINDOW_SIZE;
-    private int maxConcurrentStreams = -1;
     private int sendMaxFrameSize = DEFAULT_MAX_FRAME_SIZE;
     private int receiveMaxFrameSize = DEFAULT_MAX_FRAME_SIZE;
     private int unackedReceiveMaxFrameSize = DEFAULT_MAX_FRAME_SIZE; //the old max frame size, this gets updated when our setting frame is acked
-    private int maxHeaderListSize = -1;
 
     /**
      * How much data we have told the remote endpoint we are prepared to accept.
@@ -138,7 +146,7 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
     /**
      * How much data we can send to the remote endpoint, at the connection level.
      */
-    private volatile int sendWindowSize = initialSendWindowSize;
+    private volatile long sendWindowSize = initialSendWindowSize;
 
     private boolean thisGoneAway = false;
     private boolean peerGoneAway = false;
@@ -149,6 +157,8 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
 
     private final HpackDecoder decoder;
     private final HpackEncoder encoder;
+    private final int maxPadding;
+    private final Random paddingRandom;
 
     private int prefaceCount;
     private boolean initialSettingsReceived; //settings frame must be the first frame we relieve
@@ -163,8 +173,6 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
 
     private final Map<AttachmentKey<?>, Object> attachments = Collections.synchronizedMap(new HashMap<AttachmentKey<?>, Object>());
 
-    private final Http2PriorityTree priorityTree;
-
 
     public Http2Channel(StreamConnection connectedStreamChannel, String protocol, ByteBufferPool bufferPool, PooledByteBuffer data, boolean clientSide, boolean fromUpgrade, OptionMap settings) {
         this(connectedStreamChannel, protocol, bufferPool, data, clientSide, fromUpgrade, true, null, settings);
@@ -176,11 +184,20 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
     public Http2Channel(StreamConnection connectedStreamChannel, String protocol, ByteBufferPool bufferPool, PooledByteBuffer data, boolean clientSide, boolean fromUpgrade, boolean prefaceRequired, ByteBuffer initialOtherSideSettings, OptionMap settings) {
         super(connectedStreamChannel, bufferPool, Http2FramePriority.INSTANCE, data, settings);
         streamIdCounter = clientSide ? (fromUpgrade ? 3 : 1) : 2;
+
         pushEnabled = settings.get(UndertowOptions.HTTP2_SETTINGS_ENABLE_PUSH, true);
+        this.initialReceiveWindowSize = settings.get(UndertowOptions.HTTP2_SETTINGS_INITIAL_WINDOW_SIZE, DEFAULT_INITIAL_WINDOW_SIZE);
+
         this.protocol = protocol == null ? Http2OpenListener.HTTP2 : protocol;
 
         encoderHeaderTableSize = settings.get(UndertowOptions.HTTP2_SETTINGS_HEADER_TABLE_SIZE, Hpack.DEFAULT_TABLE_SIZE);
         receiveMaxFrameSize = settings.get(UndertowOptions.HTTP2_SETTINGS_MAX_FRAME_SIZE, DEFAULT_MAX_FRAME_SIZE);
+        maxPadding = settings.get(UndertowOptions.HTTP2_PADDING_SIZE, 0);
+        if(maxPadding > 0) {
+            paddingRandom = new SecureRandom();
+        } else {
+            paddingRandom = null;
+        }
 
         this.decoder = new HpackDecoder(Hpack.DEFAULT_TABLE_SIZE);
         this.encoder = new HpackEncoder(encoderHeaderTableSize);
@@ -193,11 +210,15 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
             prefaceCount = PREFACE_BYTES.length;
             sendSettings();
             initialSettingsSent = true;
+            if(fromUpgrade) {
+                StreamHolder streamHolder = new StreamHolder((Http2StreamSinkChannel) null);
+                streamHolder.sinkClosed = true;
+                currentStreams.put(1, streamHolder);
+            }
         } else if(fromUpgrade) {
             sendSettings();
             initialSettingsSent = true;
         }
-        priorityTree = clientSide ? null : new Http2PriorityTree();
         if (initialOtherSideSettings != null) {
             Http2SettingsParser parser = new Http2SettingsParser(initialOtherSideSettings.remaining());
             try {
@@ -220,8 +241,9 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
             settings.add(new Http2Setting(Http2Setting.SETTINGS_ENABLE_PUSH, pushEnabled ? 1 : 0));
         }
         settings.add(new Http2Setting(Http2Setting.SETTINGS_MAX_FRAME_SIZE, receiveMaxFrameSize));
+        settings.add(new Http2Setting(Http2Setting.SETTINGS_INITIAL_WINDOW_SIZE, initialReceiveWindowSize));
         Http2SettingsStreamSinkChannel stream = new Http2SettingsStreamSinkChannel(this, settings);
-        flushChannel(stream);
+        flushChannelIgnoreFailure(stream);
     }
 
     private void sendSettingsAck() {
@@ -230,28 +252,32 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
             initialSettingsSent = true;
         }
         Http2SettingsStreamSinkChannel stream = new Http2SettingsStreamSinkChannel(this);
-        flushChannel(stream);
+        flushChannelIgnoreFailure(stream);
     }
 
-    private void flushChannel(StreamSinkChannel stream) {
+    private void flushChannelIgnoreFailure(StreamSinkChannel stream) {
         try {
-            stream.shutdownWrites();
-            if (!stream.flush()) {
-                stream.getWriteSetter().set(ChannelListeners.flushingChannelListener(null, writeExceptionHandler()));
-                stream.resumeWrites();
-            }
+            flushChannel(stream);
         } catch (IOException e) {
-            //the channel excption handling will close the channel
             UndertowLogger.REQUEST_IO_LOGGER.ioException(e);
         }
     }
 
-    private void sendPreface() {
-        Http2PrefaceStreamSinkChannel preface = new Http2PrefaceStreamSinkChannel(this);
-        flushChannel(preface);
+    private void flushChannel(StreamSinkChannel stream) throws IOException {
+        stream.shutdownWrites();
+        if (!stream.flush()) {
+            stream.getWriteSetter().set(ChannelListeners.flushingChannelListener(null, writeExceptionHandler()));
+            stream.resumeWrites();
+        }
     }
 
 
+
+    private void sendPreface() {
+        Http2PrefaceStreamSinkChannel preface = new Http2PrefaceStreamSinkChannel(this);
+        flushChannelIgnoreFailure(preface);
+    }
+
     @Override
     protected AbstractHttp2StreamSourceChannel createChannel(FrameHeaderData frameHeaderData, PooledByteBuffer frameData) throws IOException {
 
@@ -259,7 +285,7 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
         AbstractHttp2StreamSourceChannel channel;
         if (frameParser.type == FRAME_TYPE_DATA) {
             //DATA frames must be already associated with a connection. If it gets here then something is wrong
-            if (frameParser.streamId == 0) {
+            if (frameParser.streamId == 0 || isIdle(frameParser.streamId)) {
                 //spec explicitly calls this out as a connection error
                 sendGoAway(ERROR_PROTOCOL_ERROR);
             } else {
@@ -272,8 +298,9 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
         //note that not all frame types are covered here, as some are only relevant to already active streams
         //if which case they are handled by the existing channel support
         switch (frameParser.type) {
-            case FRAME_TYPE_PUSH_PROMISE:
-            case FRAME_TYPE_CONTINUATION: {
+
+            case FRAME_TYPE_CONTINUATION:
+            case FRAME_TYPE_PUSH_PROMISE: {
                 //this is some 'clever' code to deal with both types continuation (push_promise and headers)
                 //if the continuation is not a push promise it falls through to the headers code
                 if(frameParser.parser instanceof Http2PushPromiseParser) {
@@ -288,27 +315,62 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
                 //fall through
             }
             case FRAME_TYPE_HEADERS: {
+                if(!isIdle(frameParser.streamId)) {
+                    //this is an existing stream
+                    //make sure it exists
+                    StreamHolder existing = currentStreams.get(frameParser.streamId);
+                    if(existing == null || existing.sourceClosed) {
+                        sendRstStream(frameParser.streamId, ERROR_STREAM_CLOSED);
+                        frameData.close();
+                        return null;
+                    } else if (existing.sourceChannel != null ){
+                        //if exists
+                        //make sure it has END_STREAM set
+                        if(!Bits.allAreSet(frameParser.flags, HEADERS_FLAG_END_STREAM)) {
+                            sendGoAway(ERROR_PROTOCOL_ERROR);
+                            frameData.close();
+                            return null;
+                        }
+                    }
+                } else {
+                    if(frameParser.streamId % 2 == (isClient() ? 1 : 0)) {
+                        sendGoAway(ERROR_PROTOCOL_ERROR);
+                        frameData.close();
+                        return null;
+                    }
+                }
                 Http2HeadersParser parser = (Http2HeadersParser) frameParser.parser;
 
                 channel = new Http2StreamSourceChannel(this, frameData, frameHeaderData.getFrameLength(), parser.getHeaderMap(), frameParser.streamId);
                 lastGoodStreamId = Math.max(lastGoodStreamId, frameParser.streamId);
+
+                StreamHolder holder = currentStreams.get(frameParser.streamId);
+                if(holder == null) {
+                    currentStreams.put(frameParser.streamId, holder = new StreamHolder((Http2StreamSourceChannel) channel));
+                } else {
+                    holder.sourceChannel = (Http2StreamSourceChannel) channel;
+                }
                 if (parser.isHeadersEndStream() && Bits.allAreSet(frameParser.flags, HEADERS_FLAG_END_HEADERS)) {
                     channel.lastFrame();
-                } else {
-                    incomingStreams.put(frameParser.streamId, (Http2StreamSourceChannel) channel);
+                    holder.sourceChannel = null;
+                    //this is yuck
+                    if(!isClient() || !"100".equals(parser.getHeaderMap().getFirst(STATUS))) {
+                        holder.sourceClosed = true;
+                    }
                 }
                 if(parser.isInvalid()) {
                     channel.rstStream(ERROR_PROTOCOL_ERROR);
-                    sendRstStream(frameParser.streamId, Http2Channel.ERROR_CANCEL);
+                    sendRstStream(frameParser.streamId, Http2Channel.ERROR_PROTOCOL_ERROR);
                     channel = null;
                 }
                 if(parser.getDependentStreamId() == frameParser.streamId) {
                     sendRstStream(frameParser.streamId, ERROR_PROTOCOL_ERROR);
+                    frameData.close();
                     return null;
                 }
-                if(priorityTree != null) {
-                    priorityTree.registerStream(frameParser.streamId, parser.getDependentStreamId(), parser.getWeight(), parser.isExclusive());
-                }
+//                if(priorityTree != null) {
+//                    priorityTree.registerStream(frameParser.streamId, parser.getDependentStreamId(), parser.getWeight(), parser.isExclusive());
+//                }
                 break;
             }
             case FRAME_TYPE_RST_STREAM: {
@@ -321,12 +383,20 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
                 }
                 channel = new Http2RstStreamStreamSourceChannel(this, frameData, parser.getErrorCode(), frameParser.streamId);
                 handleRstStream(frameParser.streamId);
+                if(isIdle(frameParser.streamId)) {
+                    sendGoAway(ERROR_PROTOCOL_ERROR);
+                }
                 break;
             }
             case FRAME_TYPE_SETTINGS: {
                 if (!Bits.anyAreSet(frameParser.flags, SETTINGS_FLAG_ACK)) {
-                    updateSettings(((Http2SettingsParser) frameParser.parser).getSettings());
-                    sendSettingsAck();
+                    if(updateSettings(((Http2SettingsParser) frameParser.parser).getSettings())) {
+                        sendSettingsAck();
+                    }
+                } else if (frameHeaderData.getFrameLength() != 0) {
+                    sendGoAway(ERROR_FRAME_SIZE_ERROR);
+                    frameData.close();
+                    return null;
                 }
                 channel = new Http2SettingsStreamSourceChannel(this, frameData, frameParser.getFrameLength(), ((Http2SettingsParser) frameParser.parser).getSettings());
                 unackedReceiveMaxFrameSize = receiveMaxFrameSize;
@@ -343,17 +413,20 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
                 break;
             }
             case FRAME_TYPE_GOAWAY: {
-                Http2GoAwayParser spdyGoAwayParser = (Http2GoAwayParser) frameParser.parser;
-                channel = new Http2GoAwayStreamSourceChannel(this, frameData, frameParser.getFrameLength(), spdyGoAwayParser.getStatusCode(), spdyGoAwayParser.getLastGoodStreamId());
+                Http2GoAwayParser http2GoAwayParser = (Http2GoAwayParser) frameParser.parser;
+                channel = new Http2GoAwayStreamSourceChannel(this, frameData, frameParser.getFrameLength(), http2GoAwayParser.getStatusCode(), http2GoAwayParser.getLastGoodStreamId());
                 peerGoneAway = true;
                 //the peer is going away
                 //everything is broken
-                for(Http2StreamSourceChannel stream : incomingStreams.values()) {
-                    stream.close();
-                }
-                for(Http2StreamSinkChannel stream : outgoingStreams.values()) {
-                    stream.close();
+                for(StreamHolder holder : currentStreams.values()) {
+                    if(holder.sourceChannel != null) {
+                        holder.sourceChannel.rstStream();
+                    }
+                    if(holder.sinkChannel != null) {
+                        holder.sinkChannel.rstStream();
+                    }
                 }
+                frameData.close();
                 break;
             }
             case FRAME_TYPE_WINDOW_UPDATE: {
@@ -371,17 +444,18 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
                     return null;
                 }
                 frameData.close();
-                if(priorityTree == null) {
-                    //we don't care, because we are the client side
-                    //so this situation should never happen
-                    return null;
-                }
-                priorityTree.priorityFrame(frameParser.streamId, parser.getStreamDependency(), parser.getWeight(), parser.isExclusive());
-                //we don't return priority notifications, they are handled internally
+//                if(priorityTree == null) {
+//                    //we don't care, because we are the client side
+//                    //so this situation should never happen
+//                    return null;
+//                }
+//                priorityTree.priorityFrame(frameParser.streamId, parser.getStreamDependency(), parser.getWeight(), parser.isExclusive());
+//                //we don't return priority notifications, they are handled internally
                 return null;
             }
             default: {
                 UndertowLogger.REQUEST_LOGGER.tracef("Dropping frame of length %s and type %s for stream %s as we do not understand this type of frame", frameParser.getFrameLength(), frameParser.type, frameParser.streamId);
+                frameData.close();
                 return null;
             }
         }
@@ -410,11 +484,11 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
         if (!initialSettingsReceived) {
             if (frameParser.type != FRAME_TYPE_SETTINGS) {
                 UndertowLogger.REQUEST_IO_LOGGER.remoteEndpointFailedToSendInitialSettings(frameParser.type);
-                StringBuilder sb = new StringBuilder();
-                while (data.hasRemaining()) {
-                    sb.append(data.get());
-                    sb.append(" ");
-                }
+                //StringBuilder sb = new StringBuilder();
+                //while (data.hasRemaining()) {
+                //    sb.append((char)data.get());
+                //    sb.append(" ");
+                //}
                 markReadsBroken(new IOException());
             } else {
                 initialSettingsReceived = true;
@@ -438,7 +512,7 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
         if(!peerGoneAway && !thisGoneAway) {
             //the peer has performed an unclean close
             //if they have streams that are still expecting data then this is an error condition
-            if(incomingStreams.size() > 0) {
+            if(currentStreams.size() > 0) {
                 //we assume something happened to the underlying connection
                 //we attempt to send our own GOAWAY, however it will probably fail,
                 //which will trigger a forces close of our write side
@@ -482,23 +556,24 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
     @Override
     protected void closeSubChannels() {
 
-        for (Map.Entry<Integer, Http2StreamSourceChannel> e : incomingStreams.entrySet()) {
-            AbstractHttp2StreamSourceChannel receiver = e.getValue();
-            if (receiver.isReadResumed()) {
-                ChannelListeners.invokeChannelListener(receiver.getIoThread(), receiver, ((ChannelListener.SimpleSetter) receiver.getReadSetter()).get());
+        for (Map.Entry<Integer, StreamHolder> e : currentStreams.entrySet()) {
+            StreamHolder holder = e.getValue();
+            AbstractHttp2StreamSourceChannel receiver = holder.sourceChannel;
+            if(receiver != null) {
+                if (receiver.isReadResumed()) {
+                    ChannelListeners.invokeChannelListener(receiver.getIoThread(), receiver, ((ChannelListener.SimpleSetter) receiver.getReadSetter()).get());
+                }
+                IoUtils.safeClose(receiver);
+            }
+            Http2StreamSinkChannel sink = holder.sinkChannel;
+            if(sink != null) {
+                if (sink.isWritesShutdown()) {
+                    ChannelListeners.invokeChannelListener(sink.getIoThread(), sink, ((ChannelListener.SimpleSetter) sink.getWriteSetter()).get());
+                }
+                IoUtils.safeClose(sink);
             }
-            IoUtils.safeClose(receiver);
-        }
-        incomingStreams.clear();
 
-        for (Map.Entry<Integer, Http2StreamSinkChannel> e : outgoingStreams.entrySet()) {
-            Http2StreamSinkChannel receiver = e.getValue();
-            if (receiver.isWritesShutdown()) {
-                ChannelListeners.invokeChannelListener(receiver.getIoThread(), receiver, ((ChannelListener.SimpleSetter) receiver.getWriteSetter()).get());
-            }
-            IoUtils.safeClose(receiver);
         }
-        outgoingStreams.clear();
     }
 
     /**
@@ -506,25 +581,29 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
      *
      * @param settings
      */
-    synchronized void updateSettings(List<Http2Setting> settings) {
+    synchronized boolean updateSettings(List<Http2Setting> settings) {
         for (Http2Setting setting : settings) {
             if (setting.getId() == Http2Setting.SETTINGS_INITIAL_WINDOW_SIZE) {
                 int old = initialSendWindowSize;
-                initialSendWindowSize = setting.getValue();
+                if(setting.getValue() > Integer.MAX_VALUE) {
+                    sendGoAway(ERROR_FLOW_CONTROL_ERROR);
+                    return false;
+                }
+                initialSendWindowSize = (int) setting.getValue();
                 int difference = initialSendWindowSize - old;
                 sendWindowSize += difference;
             } else if (setting.getId() == Http2Setting.SETTINGS_MAX_FRAME_SIZE) {
-                if(sendMaxFrameSize > MAX_FRAME_SIZE) {
+                if(setting.getValue() > MAX_FRAME_SIZE || setting.getValue() < DEFAULT_MAX_FRAME_SIZE) {
                     UndertowLogger.REQUEST_IO_LOGGER.debug("Invalid value received for SETTINGS_MAX_FRAME_SIZE " + setting.getValue());
                     sendGoAway(ERROR_PROTOCOL_ERROR);
-                    return;
+                    return false;
                 }
-                sendMaxFrameSize = setting.getValue();
+                sendMaxFrameSize = (int) setting.getValue();
             } else if (setting.getId() == Http2Setting.SETTINGS_HEADER_TABLE_SIZE) {
-                encoder.setMaxTableSize(setting.getValue());
+                encoder.setMaxTableSize((int) setting.getValue());
             } else if (setting.getId() == Http2Setting.SETTINGS_ENABLE_PUSH) {
 
-                int result = setting.getValue();
+                int result = (int) setting.getValue();
                 //we allow the remote endpoint to disable push
                 //but not enable it if it has been explictly disabled on this side
                 if(result == 0) {
@@ -533,11 +612,12 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
                     //invalid value
                     UndertowLogger.REQUEST_IO_LOGGER.debug("Invalid value received for SETTINGS_ENABLE_PUSH " + result);
                     sendGoAway(ERROR_PROTOCOL_ERROR);
-                    return;
+                    return false;
                 }
             }
             //ignore the rest for now
         }
+        return true;
     }
 
     public int getHttp2Version() {
@@ -554,15 +634,33 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
 
     public synchronized void handleWindowUpdate(int streamId, int deltaWindowSize) throws IOException {
         if (streamId == 0) {
-            boolean exhausted = sendWindowSize == 0;
+            if (deltaWindowSize == 0) {
+                UndertowLogger.REQUEST_IO_LOGGER.debug("Invalid flow-control window increment of 0 received with WINDOW_UPDATE frame for connection");
+                sendGoAway(ERROR_PROTOCOL_ERROR);
+                return;
+            }
+
+            boolean exhausted = sendWindowSize <= FLOW_CONTROL_MIN_WINDOW; //
+
             sendWindowSize += deltaWindowSize;
             if (exhausted) {
                 notifyFlowControlAllowed();
             }
+            if(sendWindowSize > Integer.MAX_VALUE) {
+                sendGoAway(ERROR_FLOW_CONTROL_ERROR);
+            }
         } else {
-            Http2StreamSinkChannel stream = outgoingStreams.get(streamId);
+            if (deltaWindowSize == 0) {
+                UndertowLogger.REQUEST_IO_LOGGER.debug("Invalid flow-control window increment of 0 received with WINDOW_UPDATE frame for stream " + streamId);
+                sendRstStream(streamId, ERROR_PROTOCOL_ERROR);
+                return;
+            }
+            StreamHolder holder = currentStreams.get(streamId);
+            Http2StreamSinkChannel stream = holder != null ? holder.sinkChannel : null;
             if (stream == null) {
-                //TODO: error handling
+                if(isIdle(streamId)) {
+                    sendGoAway(ERROR_PROTOCOL_ERROR);
+                }
             } else {
                 stream.updateFlowControlWindow(deltaWindowSize);
             }
@@ -622,7 +720,7 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
         }
     }
 
-    public void sendUpdateWindowSize(int streamId, int delta) {
+    public void sendUpdateWindowSize(int streamId, int delta) throws IOException {
         Http2WindowUpdateStreamSinkChannel windowUpdateStreamSinkChannel = new Http2WindowUpdateStreamSinkChannel(this, streamId, delta);
         flushChannel(windowUpdateStreamSinkChannel);
 
@@ -636,7 +734,7 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
         return null;
     }
 
-    public synchronized void updateReceiveFlowControlWindow(int read) {
+    public synchronized void updateReceiveFlowControlWindow(int read) throws IOException {
         if (read <= 0) {
             return;
         }
@@ -666,9 +764,9 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
         }
         int streamId = streamIdCounter;
         streamIdCounter += 2;
-        Http2HeadersStreamSinkChannel spdySynStreamStreamSinkChannel = new Http2HeadersStreamSinkChannel(this, streamId, requestHeaders);
-        outgoingStreams.put(streamId, spdySynStreamStreamSinkChannel);
-        return spdySynStreamStreamSinkChannel;
+        Http2HeadersStreamSinkChannel http2SynStreamStreamSinkChannel = new Http2HeadersStreamSinkChannel(this, streamId, requestHeaders);
+        currentStreams.put(streamId, new StreamHolder(http2SynStreamStreamSinkChannel));
+        return http2SynStreamStreamSinkChannel;
     }
 
     public synchronized Http2HeadersStreamSinkChannel sendPushPromise(int associatedStreamId, HeaderMap requestHeaders, HeaderMap responseHeaders) throws IOException {
@@ -683,9 +781,9 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
         Http2PushPromiseStreamSinkChannel pushPromise = new Http2PushPromiseStreamSinkChannel(this, requestHeaders, associatedStreamId, streamId);
         flushChannel(pushPromise);
 
-        Http2HeadersStreamSinkChannel spdySynStreamStreamSinkChannel = new Http2HeadersStreamSinkChannel(this, streamId, responseHeaders);
-        outgoingStreams.put(streamId, spdySynStreamStreamSinkChannel);
-        return spdySynStreamStreamSinkChannel;
+        Http2HeadersStreamSinkChannel http2SynStreamStreamSinkChannel = new Http2HeadersStreamSinkChannel(this, streamId, responseHeaders);
+        currentStreams.put(streamId, new StreamHolder(http2SynStreamStreamSinkChannel));
+        return http2SynStreamStreamSinkChannel;
     }
 
     /**
@@ -695,27 +793,39 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
      * @return The actual amount of bytes the sender can send
      */
     synchronized int grabFlowControlBytes(int bytesToGrab) {
-        int min = Math.min(bytesToGrab, sendWindowSize);
+        int min = (int) Math.min(bytesToGrab, sendWindowSize);
+        if(bytesToGrab > FLOW_CONTROL_MIN_WINDOW && min <= FLOW_CONTROL_MIN_WINDOW) {
+            //this can cause problems with padding, so we just return 0
+            return 0;
+        }
         min = Math.min(sendMaxFrameSize, min);
         sendWindowSize -= min;
         return min;
     }
 
     void registerStreamSink(Http2HeadersStreamSinkChannel synResponse) {
-        outgoingStreams.put(synResponse.getStreamId(), synResponse);
+        StreamHolder existing = currentStreams.get(synResponse.getStreamId());
+        if(existing == null) {
+            throw UndertowMessages.MESSAGES.streamNotRegistered();
+        }
+        existing.sinkChannel = synResponse;
     }
 
     void removeStreamSink(int streamId) {
-        outgoingStreams.remove(streamId);
-        if(isLastFrameReceived() && outgoingStreams.isEmpty()) {
+        StreamHolder existing = currentStreams.get(streamId);
+        if(existing == null) {
+            return;
+        }
+        existing.sinkClosed = true;
+        existing.sinkChannel = null;
+        if(existing.sourceClosed) {
+            currentStreams.remove(streamId);
+        }
+        if(isLastFrameReceived() && currentStreams.isEmpty()) {
             sendGoAway(ERROR_NO_ERROR);
         }
     }
 
-    Map<Integer, Http2StreamSourceChannel> getIncomingStreams() {
-        return incomingStreams;
-    }
-
     public boolean isClient() {
         return streamIdCounter % 2 == 1;
     }
@@ -729,6 +839,13 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
         return decoder;
     }
 
+    int getPaddingBytes() {
+        if(paddingRandom == null) {
+            return 0;
+        }
+        return paddingRandom.nextInt(maxPadding);
+    }
+
     @Override
     public <T> T getAttachment(AttachmentKey<T> key) {
         if (key == null) {
@@ -784,17 +901,18 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
     public void sendRstStream(int streamId, int statusCode) {
         handleRstStream(streamId);
         Http2RstStreamSinkChannel channel = new Http2RstStreamSinkChannel(this, streamId, statusCode);
-        flushChannel(channel);
+        flushChannelIgnoreFailure(channel);
     }
 
     private void handleRstStream(int streamId) {
-        AbstractHttp2StreamSourceChannel incoming = incomingStreams.remove(streamId);
-        if (incoming != null) {
-            incoming.rstStream();
-        }
-        Http2StreamSinkChannel outgoing = outgoingStreams.remove(streamId);
-        if (outgoing != null) {
-            outgoing.rstStream();
+        StreamHolder holder = currentStreams.remove(streamId);
+        if(holder != null) {
+            if (holder.sinkChannel != null) {
+                holder.sinkChannel.rstStream();
+            }
+            if (holder.sourceChannel != null) {
+                holder.sourceChannel.rstStream();
+            }
         }
     }
 
@@ -809,7 +927,9 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
         }
         lastGoodStreamId = 1;
         Http2HeadersStreamSinkChannel stream = new Http2HeadersStreamSinkChannel(this, 1);
-        outgoingStreams.put(1, stream);
+        StreamHolder streamHolder = new StreamHolder(stream);
+        streamHolder.sourceClosed = true;
+        currentStreams.put(1, streamHolder);
         return stream;
 
     }
@@ -826,6 +946,28 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
         return thisGoneAway;
     }
 
+    Http2StreamSourceChannel removeStreamSource(int streamId) {
+        StreamHolder existing = currentStreams.get(streamId);
+        if(existing == null){
+            return null;
+        }
+        existing.sourceClosed = true;
+        Http2StreamSourceChannel ret = existing.sourceChannel;
+        existing.sourceChannel = null;
+        if(existing.sinkClosed) {
+            currentStreams.remove(streamId);
+        }
+        return ret;
+    }
+
+    Http2StreamSourceChannel getIncomingStream(int streamId) {
+        StreamHolder existing = currentStreams.get(streamId);
+        if(existing == null){
+            return null;
+        }
+        return existing.sourceChannel;
+    }
+
     private class Http2ControlMessageExceptionHandler implements ChannelExceptionHandler<AbstractHttp2StreamSinkChannel> {
         @Override
         public void handleException(AbstractHttp2StreamSinkChannel channel, IOException exception) {
@@ -845,4 +987,27 @@ public class Http2Channel extends AbstractFramedChannel<Http2Channel, AbstractHt
     public String getProtocol() {
         return protocol;
     }
+
+    private boolean isIdle(int streamNo) {
+        if(streamNo % 2 == streamIdCounter % 2) {
+            return streamNo >= streamIdCounter;
+        } else {
+            return streamNo > lastGoodStreamId;
+        }
+    }
+
+    static final class StreamHolder {
+        boolean sourceClosed = false;
+        boolean sinkClosed = false;
+        Http2StreamSourceChannel sourceChannel;
+        Http2StreamSinkChannel sinkChannel;
+
+        StreamHolder(Http2StreamSourceChannel sourceChannel) {
+            this.sourceChannel = sourceChannel;
+        }
+
+        StreamHolder(Http2StreamSinkChannel sinkChannel) {
+            this.sinkChannel = sinkChannel;
+        }
+    }
 }
