@@ -62,6 +62,8 @@ import static org.xnio.Bits.anyAreSet;
  */
 public class SslConduit implements StreamSourceConduit, StreamSinkConduit {
 
+    public static final int MAX_READ_LISTENER_INVOCATIONS = Integer.getInteger("io.undertow.ssl.max-read-listener-invocations", 100);
+
     /**
      * If this is set we are in the middle of a handshake, and we cannot
      * read any more data until we have written out our wrap result
@@ -118,13 +120,13 @@ public class SslConduit implements StreamSourceConduit, StreamSinkConduit {
 
     private final UndertowSslConnection connection;
     private final StreamConnection delegate;
-    private final SSLEngine engine;
+    private SSLEngine engine;
     private final StreamSinkConduit sink;
     private final StreamSourceConduit source;
     private final ByteBufferPool bufferPool;
     private final Runnable handshakeCallback;
 
-    private int state = 0;
+    private volatile int state = 0;
 
     private volatile int outstandingTasks = 0;
 
@@ -152,9 +154,36 @@ public class SslConduit implements StreamSourceConduit, StreamSinkConduit {
 
     private SslWriteReadyHandler writeReadyHandler;
     private SslReadReadyHandler readReadyHandler;
+    private int readListenerInvocationCount;
 
     private boolean invokingReadListenerHandshake = false;
 
+
+
+    private final Runnable runReadListenerCommand = new Runnable() {
+        @Override
+        public void run() {
+            final int count = readListenerInvocationCount;
+            try {
+                readReadyHandler.readReady();
+            } finally {
+                if(count == readListenerInvocationCount) {
+                    readListenerInvocationCount = 0;
+                }
+            }
+        }
+    };
+
+    private final Runnable runReadListenerAndResumeCommand = new Runnable() {
+        @Override
+        public void run() {
+            if (allAreSet(state, FLAG_READS_RESUMED)) {
+                delegate.getSourceChannel().resumeReads();
+            }
+            runReadListenerCommand.run();
+        }
+    };
+
     SslConduit(UndertowSslConnection connection, StreamConnection delegate, SSLEngine engine, ByteBufferPool bufferPool, Runnable handshakeCallback) {
         this.connection = connection;
         this.delegate = delegate;
@@ -220,15 +249,17 @@ public class SslConduit implements StreamSourceConduit, StreamSinkConduit {
 
     private void runReadListener(final boolean resumeInListener) {
         try {
-            delegate.getIoThread().execute(new Runnable() {
-                @Override
-                public void run() {
-                    if(resumeInListener) {
-                        delegate.getSourceChannel().resumeReads();
-                    }
-                    readReadyHandler.readReady();
-                }
-            });
+            if(readListenerInvocationCount++ == MAX_READ_LISTENER_INVOCATIONS) {
+                UndertowLogger.REQUEST_LOGGER.sslReadLoopDetected(this);
+                IoUtils.safeClose(connection, delegate);
+                close();
+                return;
+            }
+            if(resumeInListener) {
+                delegate.getIoThread().execute(runReadListenerAndResumeCommand);
+            } else {
+                delegate.getIoThread().execute(runReadListenerCommand);
+            }
         } catch (Exception e) {
             //will only happen on shutdown
             IoUtils.safeClose(connection, delegate);
@@ -501,7 +532,13 @@ public class SslConduit implements StreamSourceConduit, StreamSinkConduit {
             if(allAreClear(state, FLAG_DELEGATE_SINK_SHUTDOWN)) {
                 sink.terminateWrites();
                 state |= FLAG_DELEGATE_SINK_SHUTDOWN;
+                notifyWriteClosed();
             }
+            boolean result = sink.flush();
+            if(result && anyAreSet(state, FLAG_READ_CLOSED)) {
+                closed();
+            }
+            return result;
         }
         return sink.flush();
     }
@@ -509,7 +546,7 @@ public class SslConduit implements StreamSourceConduit, StreamSinkConduit {
     @Override
     public long transferTo(long position, long count, FileChannel target) throws IOException {
         if(anyAreSet(state, FLAG_READ_SHUTDOWN)) {
-            throw new ClosedChannelException();
+            return -1;
         }
         return target.transferFrom(new ConduitReadableByteChannel(this), position, count);
     }
@@ -517,7 +554,7 @@ public class SslConduit implements StreamSourceConduit, StreamSinkConduit {
     @Override
     public long transferTo(long count, ByteBuffer throughBuffer, StreamSinkChannel target) throws IOException {
         if(anyAreSet(state, FLAG_READ_SHUTDOWN)) {
-            throw new ClosedChannelException();
+            return -1;
         }
         return IoUtils.transfer(new ConduitReadableByteChannel(this), count, throughBuffer, target);
     }
@@ -525,7 +562,7 @@ public class SslConduit implements StreamSourceConduit, StreamSinkConduit {
     @Override
     public int read(ByteBuffer dst) throws IOException {
         if(anyAreSet(state, FLAG_READ_SHUTDOWN)) {
-            throw new ClosedChannelException();
+            return -1;
         }
         return (int) doUnwrap(new ByteBuffer[]{dst}, 0, 1);
     }
@@ -533,7 +570,7 @@ public class SslConduit implements StreamSourceConduit, StreamSinkConduit {
     @Override
     public long read(ByteBuffer[] dsts, int offs, int len) throws IOException {
         if(anyAreSet(state, FLAG_READ_SHUTDOWN)) {
-            throw new ClosedChannelException();
+            return -1;
         }
         return doUnwrap(dsts, offs, len);
     }
@@ -577,7 +614,7 @@ public class SslConduit implements StreamSourceConduit, StreamSinkConduit {
             UndertowLogger.REQUEST_IO_LOGGER.ioException(new IOException(e));
         }
 
-        state |= FLAG_READ_CLOSED | FLAG_ENGINE_INBOUND_SHUTDOWN;
+        state |= FLAG_READ_CLOSED | FLAG_ENGINE_INBOUND_SHUTDOWN | FLAG_READ_SHUTDOWN;
         if(anyAreSet(state, FLAG_WRITE_CLOSED)) {
             closed();
         }
@@ -633,24 +670,29 @@ public class SslConduit implements StreamSourceConduit, StreamSinkConduit {
                 return 0;
             }
         }
-
+        boolean bytesProduced = false;
         PooledByteBuffer unwrappedData = this.unwrappedData;
         //copy any exiting data
-        if(unwrappedData != null && userBuffers != null) {
-            long copied = Buffers.copy(userBuffers, off, len, unwrappedData.getBuffer());
-            if(!unwrappedData.getBuffer().hasRemaining()) {
-                unwrappedData.close();
-                this.unwrappedData = null;
+        if(unwrappedData != null) {
+            if(userBuffers != null) {
+                long copied = Buffers.copy(userBuffers, off, len, unwrappedData.getBuffer());
+                if (!unwrappedData.getBuffer().hasRemaining()) {
+                    unwrappedData.close();
+                    this.unwrappedData = null;
+                }
+                if(copied > 0) {
+                    readListenerInvocationCount = 0;
+                }
+                return copied;
             }
-            return copied;
         }
         try {
             //we need to store how much data is in the unwrap buffer. If no progress can be made then we unset
             //the data to unwrap flag
-            int dataToUnwrapLength = -1;
+            int dataToUnwrapLength;
             //try and read some data if we don't already have some
-            if(allAreClear(state, FLAG_DATA_TO_UNWRAP)) {
-                if(dataToUnwrap == null) {
+            if (allAreClear(state, FLAG_DATA_TO_UNWRAP)) {
+                if (dataToUnwrap == null) {
                     dataToUnwrap = bufferPool.allocate();
                 }
                 int res;
@@ -662,19 +704,19 @@ public class SslConduit implements StreamSourceConduit, StreamSinkConduit {
                     throw e;
                 }
                 dataToUnwrap.getBuffer().flip();
-                if(res == -1) {
+                if (res == -1) {
                     dataToUnwrap.close();
                     dataToUnwrap = null;
                     notifyReadClosed();
                     return -1;
-                } else if(res == 0 && engine.getHandshakeStatus() == SSLEngineResult.HandshakeStatus.FINISHED) {
+                } else if (res == 0 && engine.getHandshakeStatus() == SSLEngineResult.HandshakeStatus.FINISHED) {
                     return 0;
                 }
             }
             dataToUnwrapLength = dataToUnwrap.getBuffer().remaining();
 
             long original = 0;
-            if(userBuffers != null) {
+            if (userBuffers != null) {
                 original = Buffers.remaining(userBuffers);
             }
             //perform the actual unwrap operation
@@ -695,6 +737,7 @@ public class SslConduit implements StreamSourceConduit, StreamSinkConduit {
                         result = engine.unwrap(this.dataToUnwrap.getBuffer(), d);
                         unwrapBufferUsed = true;
                     }
+                    bytesProduced = result.bytesProduced() > 0;
                 } else {
                     unwrapBufferUsed = true;
                     if (unwrappedData == null) {
@@ -703,11 +746,12 @@ public class SslConduit implements StreamSourceConduit, StreamSinkConduit {
                         unwrappedData.getBuffer().compact();
                     }
                     result = engine.unwrap(this.dataToUnwrap.getBuffer(), unwrappedData.getBuffer());
+                    bytesProduced = result.bytesProduced() > 0;
                 }
             } finally {
-                if(unwrapBufferUsed) {
+                if (unwrapBufferUsed) {
                     unwrappedData.getBuffer().flip();
-                    if(!unwrappedData.getBuffer().hasRemaining()) {
+                    if (!unwrappedData.getBuffer().hasRemaining()) {
                         unwrappedData.close();
                         unwrappedData = null;
                     }
@@ -716,41 +760,57 @@ public class SslConduit implements StreamSourceConduit, StreamSinkConduit {
             }
 
             if (!handleHandshakeResult(result)) {
-                if(this.dataToUnwrap.getBuffer().hasRemaining() && result.getStatus() != SSLEngineResult.Status.BUFFER_UNDERFLOW && dataToUnwrap.getBuffer().remaining() != dataToUnwrapLength) {
+                if (this.dataToUnwrap.getBuffer().hasRemaining() && result.getStatus() != SSLEngineResult.Status.BUFFER_UNDERFLOW && dataToUnwrap.getBuffer().remaining() != dataToUnwrapLength) {
                     state |= FLAG_DATA_TO_UNWRAP;
+                } else {
+                    state &= ~FLAG_DATA_TO_UNWRAP;
                 }
                 return 0;
             }
             if (result.getStatus() == SSLEngineResult.Status.CLOSED) {
+                if(dataToUnwrap != null) {
+                    dataToUnwrap.close();
+                    dataToUnwrap = null;
+                }
                 notifyReadClosed();
                 return -1;
             }
-            if(result.getStatus() == SSLEngineResult.Status.BUFFER_UNDERFLOW) {
+            if (result.getStatus() == SSLEngineResult.Status.BUFFER_UNDERFLOW) {
                 state &= ~FLAG_DATA_TO_UNWRAP;
-            } else if(result.getStatus() == SSLEngineResult.Status.BUFFER_OVERFLOW) {
-                throw new IOException("overflow"); //todo: handle properly
-            } else if(this.dataToUnwrap.getBuffer().hasRemaining() && dataToUnwrap.getBuffer().remaining() != dataToUnwrapLength) {
+            } else if (result.getStatus() == SSLEngineResult.Status.BUFFER_OVERFLOW) {
+                UndertowLogger.REQUEST_LOGGER.sslBufferOverflow(this);
+                IoUtils.safeClose(delegate);
+            } else if (this.dataToUnwrap.getBuffer().hasRemaining() && dataToUnwrap.getBuffer().remaining() != dataToUnwrapLength) {
                 state |= FLAG_DATA_TO_UNWRAP;
             } else {
                 state &= ~FLAG_DATA_TO_UNWRAP;
             }
-            if(userBuffers == null) {
+            if (userBuffers == null) {
                 return 0;
             } else {
-                return original - Buffers.remaining(userBuffers);
+                long res = original - Buffers.remaining(userBuffers);
+                if(res > 0) {
+                    //if data has been successfully returned this is not a read loop
+                    readListenerInvocationCount = 0;
+                }
+                return res;
             }
+        } catch (RuntimeException|IOException e) {
+            close();
+            throw e;
         } finally {
             boolean requiresListenerInvocation = false; //if there is data in the buffer and reads are resumed we should re-run the listener
-            if (unwrappedData != null && unwrappedData.getBuffer().hasRemaining()) {
+            //we always need to re-invoke if bytes have been produced, as the engine may have buffered some data
+            if (bytesProduced || (unwrappedData != null && unwrappedData.isOpen() && unwrappedData.getBuffer().hasRemaining())) {
                 requiresListenerInvocation = true;
             }
-            if(dataToUnwrap != null) {
+            if (dataToUnwrap != null) {
                 //if there is no data in the buffer we just free it
-                if(!dataToUnwrap.getBuffer().hasRemaining()) {
+                if (!dataToUnwrap.getBuffer().hasRemaining()) {
                     dataToUnwrap.close();
                     dataToUnwrap = null;
                     state &= ~FLAG_DATA_TO_UNWRAP;
-                } else if(allAreClear(state, FLAG_DATA_TO_UNWRAP)) {
+                } else if (allAreClear(state, FLAG_DATA_TO_UNWRAP)) {
                     //if there is not enough data in the buffer we compact it to make room for more
                     dataToUnwrap.getBuffer().compact();
                 } else {
@@ -760,7 +820,7 @@ public class SslConduit implements StreamSourceConduit, StreamSinkConduit {
             }
             //if we are in the read listener handshake we don't need to invoke
             //as it is about to be invoked anyway
-            if(requiresListenerInvocation && anyAreSet(state, FLAG_READS_RESUMED) && !invokingReadListenerHandshake) {
+            if (requiresListenerInvocation && (anyAreSet(state, FLAG_READS_RESUMED) || allAreSet(state, FLAG_WRITE_REQUIRES_READ | FLAG_WRITES_RESUMED)) && !invokingReadListenerHandshake) {
                 runReadListener(false);
             }
         }
@@ -814,17 +874,17 @@ public class SslConduit implements StreamSourceConduit, StreamSinkConduit {
             if (result.getStatus() == SSLEngineResult.Status.BUFFER_UNDERFLOW) {
                 throw new IOException("underflow"); //todo: can this happen?
             } else if (result.getStatus() == SSLEngineResult.Status.BUFFER_OVERFLOW) {
-                if(!wrappedData.getBuffer().hasRemaining()) { //if an earlier wrap suceeded we ignore this
+                if (!wrappedData.getBuffer().hasRemaining()) { //if an earlier wrap suceeded we ignore this
                     throw new IOException("overflow"); //todo: handle properly
                 }
             }
             //attempt to write it out, if we fail we just return
             //we ignore the handshake status, as wrap will get called again
-            if(wrappedData.getBuffer().hasRemaining()) {
+            if (wrappedData.getBuffer().hasRemaining()) {
                 sink.write(wrappedData.getBuffer());
             }
             //if it was not a complete write we just return
-            if(wrappedData.getBuffer().hasRemaining()) {
+            if (wrappedData.getBuffer().hasRemaining()) {
                 return result.bytesConsumed();
             }
 
@@ -837,6 +897,9 @@ public class SslConduit implements StreamSourceConduit, StreamSinkConduit {
             }
 
             return result.bytesConsumed();
+        } catch (RuntimeException|IOException e) {
+            close();
+            throw e;
         } finally {
             //this can be cleared if the channel is fully closed
             if(wrappedData != null) {
@@ -1028,7 +1091,7 @@ public class SslConduit implements StreamSourceConduit, StreamSinkConduit {
 
         @Override
         public void readReady() {
-            if(anyAreSet(state, FLAG_WRITE_REQUIRES_READ) && !anyAreSet(state, FLAG_ENGINE_INBOUND_SHUTDOWN)) {
+            if(anyAreSet(state, FLAG_WRITE_REQUIRES_READ) && anyAreSet(state, FLAG_WRITES_RESUMED | FLAG_READS_RESUMED) && !anyAreSet(state, FLAG_ENGINE_INBOUND_SHUTDOWN)) {
                 try {
                     invokingReadListenerHandshake = true;
                     doHandshake();
@@ -1040,6 +1103,7 @@ public class SslConduit implements StreamSourceConduit, StreamSinkConduit {
                 }
             }
             boolean noProgress = false;
+            int initialDataToUnwrap = -1;
             int initialUnwrapped = -1;
             if (anyAreSet(state, FLAG_READS_RESUMED)) {
                 if (delegateHandler == null) {
@@ -1048,10 +1112,15 @@ public class SslConduit implements StreamSourceConduit, StreamSinkConduit {
                         suspendReads();
                     } else {
                         if(anyAreSet(state, FLAG_DATA_TO_UNWRAP)) {
-                            initialUnwrapped = dataToUnwrap.getBuffer().remaining();
+                            initialDataToUnwrap = dataToUnwrap.getBuffer().remaining();
+                        }
+                        if(unwrappedData != null) {
+                            initialUnwrapped = unwrappedData.getBuffer().remaining();
                         }
                         ChannelListeners.invokeChannelListener(connection.getSourceChannel(), readListener);
-                        if(anyAreSet(state, FLAG_DATA_TO_UNWRAP) && initialUnwrapped == dataToUnwrap.getBuffer().remaining()) {
+                        if(anyAreSet(state, FLAG_DATA_TO_UNWRAP) && initialDataToUnwrap == dataToUnwrap.getBuffer().remaining()) {
+                            noProgress = true;
+                        } else if(unwrappedData != null && unwrappedData.getBuffer().remaining() == initialUnwrapped) {
                             noProgress = true;
                         }
                     }
@@ -1059,7 +1128,7 @@ public class SslConduit implements StreamSourceConduit, StreamSinkConduit {
                     delegateHandler.readReady();
                 }
             }
-            if(!anyAreSet(state, FLAG_READS_RESUMED | FLAG_WRITE_REQUIRES_READ)) {
+            if(!anyAreSet(state, FLAG_READS_RESUMED) && !allAreSet(state, FLAG_WRITE_REQUIRES_READ | FLAG_WRITES_RESUMED)) {
                 delegate.getSourceChannel().suspendReads();
             } else if(anyAreSet(state, FLAG_READS_RESUMED) && (unwrappedData != null || anyAreSet(state, FLAG_DATA_TO_UNWRAP))) {
                 if(anyAreSet(state, FLAG_READ_CLOSED)) {
@@ -1161,6 +1230,10 @@ public class SslConduit implements StreamSourceConduit, StreamSinkConduit {
         }
     }
 
+    public void setSslEngine(SSLEngine engine) {
+        this.engine = engine;
+    }
+
     @Override
     public String toString() {
         return "SslConduit{" +
