@@ -82,7 +82,7 @@ public class ExtendedAccessLogParser {
         private boolean subToken;
         private boolean parameter;
 
-        public PatternTokenizer(String str) {
+        PatternTokenizer(String str) {
             sr = new StringReader(str);
         }
 
@@ -241,7 +241,8 @@ public class ExtendedAccessLogParser {
             if (tokenizer.hasSubToken()) {
                 String nextToken = tokenizer.getToken();
                 if ("taken".equals(nextToken)) {
-                    return new ResponseTimeAttribute(TimeUnit.SECONDS);
+                    //if response timing are not enabled we just print a '-'
+                    return new SubstituteEmptyWrapper.SubstituteEmptyAttribute(new ResponseTimeAttribute(TimeUnit.SECONDS), "-");
                 }
             } else {
                 return new DateTimeAttribute("HH:mm:ss", "GMT");
@@ -321,7 +322,7 @@ public class ExtendedAccessLogParser {
                     if ("stem".equals(token)) {
                         return RequestURLAttribute.INSTANCE;
                     } else if ("query".equals(token)) {
-                        return new SubstituteEmptyWrapper.SubstituteEmptyAttribute(QueryStringAttribute.INSTANCE, "-");
+                        return new SubstituteEmptyWrapper.SubstituteEmptyAttribute(QueryStringAttribute.BARE_INSTANCE, "-");
                     }
                 } else {
                     return new ExchangeAttribute() {
@@ -385,7 +386,7 @@ public class ExtendedAccessLogParser {
             throws IOException {
         String token = null;
         if (tokenizer.hasSubToken()) {
-            token = tokenizer.getToken();
+            tokenizer.getToken();
             return new ConstantExchangeAttribute("-");
         } else if (tokenizer.hasParameter()) {
             tokenizer.getParameter();
