@@ -18,6 +18,40 @@
 
 package io.undertow.server.handlers.proxy.mod_cluster;
 
+import io.undertow.UndertowLogger;
+import io.undertow.UndertowMessages;
+import io.undertow.Version;
+import io.undertow.io.Sender;
+import io.undertow.server.HttpHandler;
+import io.undertow.server.HttpServerExchange;
+import io.undertow.server.handlers.form.FormData;
+import io.undertow.server.handlers.form.FormDataParser;
+import io.undertow.server.handlers.form.FormEncodedDataDefinition;
+import io.undertow.server.handlers.form.FormParserFactory;
+import io.undertow.util.Headers;
+import io.undertow.util.HttpString;
+import io.undertow.util.StatusCodes;
+import org.xnio.OptionMap;
+import org.xnio.Options;
+import org.xnio.ssl.XnioSsl;
+
+import java.io.IOException;
+import java.net.InetSocketAddress;
+import java.net.URI;
+import java.net.URISyntaxException;
+import java.util.ArrayDeque;
+import java.util.ArrayList;
+import java.util.Arrays;
+import java.util.Collection;
+import java.util.Collections;
+import java.util.Deque;
+import java.util.HashSet;
+import java.util.Iterator;
+import java.util.LinkedHashMap;
+import java.util.List;
+import java.util.Map;
+import java.util.Set;
+
 import static io.undertow.server.handlers.proxy.mod_cluster.MCMPConstants.ALIAS;
 import static io.undertow.server.handlers.proxy.mod_cluster.MCMPConstants.BALANCER;
 import static io.undertow.server.handlers.proxy.mod_cluster.MCMPConstants.CONTEXT;
@@ -41,40 +75,6 @@ import static io.undertow.server.handlers.proxy.mod_cluster.MCMPConstants.TIMEOU
 import static io.undertow.server.handlers.proxy.mod_cluster.MCMPConstants.TTL;
 import static io.undertow.server.handlers.proxy.mod_cluster.MCMPConstants.TYPE;
 
-import java.io.IOException;
-import java.net.InetSocketAddress;
-import java.net.URI;
-import java.net.URISyntaxException;
-import java.util.ArrayDeque;
-import java.util.ArrayList;
-import java.util.Arrays;
-import java.util.Collection;
-import java.util.Collections;
-import java.util.Deque;
-import java.util.HashSet;
-import java.util.Iterator;
-import java.util.LinkedHashMap;
-import java.util.List;
-import java.util.Map;
-import java.util.Set;
-
-import io.undertow.UndertowLogger;
-import io.undertow.UndertowMessages;
-import io.undertow.Version;
-import io.undertow.io.Sender;
-import io.undertow.server.HttpHandler;
-import io.undertow.server.HttpServerExchange;
-import io.undertow.server.handlers.form.FormData;
-import io.undertow.server.handlers.form.FormDataParser;
-import io.undertow.server.handlers.form.FormEncodedDataDefinition;
-import io.undertow.server.handlers.form.FormParserFactory;
-import io.undertow.util.Headers;
-import io.undertow.util.HttpString;
-import io.undertow.util.StatusCodes;
-import org.xnio.OptionMap;
-import org.xnio.Options;
-import org.xnio.ssl.XnioSsl;
-
 /**
  * The mod cluster management protocol http handler.
  *
@@ -123,7 +123,7 @@ class MCMPHandler implements HttpHandler {
     private final ModCluster modCluster;
     protected final ModClusterContainer container;
 
-    public MCMPHandler(MCMPConfig config, ModCluster modCluster, HttpHandler next) {
+    MCMPHandler(MCMPConfig config, ModCluster modCluster, HttpHandler next) {
         this.config = config;
         this.next = next;
         this.modCluster = modCluster;
@@ -143,9 +143,8 @@ class MCMPHandler implements HttpHandler {
          * Proxy the request that needs to be proxied and process others
          */
         // TODO maybe this should be handled outside here?
-        final InetSocketAddress addr = exchange.getDestinationAddress();
-        //we use getHostString to avoid a reverse lookup
-        if (addr.getPort() != config.getManagementPort() || (!addr.getHostString().equals(config.getManagementHost()) &&  !addr.getHostString().equals(config.getManagementHostIp()))) {
+        final InetSocketAddress addr = exchange.getConnection().getLocalAddress(InetSocketAddress.class);
+        if (!addr.isUnresolved() && addr.getPort() != config.getManagementSocketAddress().getPort() || !Arrays.equals(addr.getAddress().getAddress(), config.getManagementSocketAddress().getAddress().getAddress())) {
             next.handleRequest(exchange);
             return;
         }
@@ -228,7 +227,7 @@ class MCMPHandler implements HttpHandler {
                 node.setBalancer(value);
                 balancer.setName(value);
             } else if (MAXATTEMPTS.equals(name)) {
-                balancer.setMaxattempts(Integer.valueOf(value));
+                balancer.setMaxattempts(Integer.parseInt(value));
             } else if (STICKYSESSION.equals(name)) {
                 if ("No".equalsIgnoreCase(value)) {
                     balancer.setStickySession(false);
@@ -264,15 +263,15 @@ class MCMPHandler implements HttpHandler {
                     node.setFlushPackets(true);
                 }
             } else if (FLUSH_WAIT.equals(name)) {
-                node.setFlushwait(Integer.valueOf(value));
+                node.setFlushwait(Integer.parseInt(value));
             } else if (MCMPConstants.PING.equals(name)) {
-                node.setPing(Integer.valueOf(value));
+                node.setPing(Integer.parseInt(value));
             } else if (SMAX.equals(name)) {
-                node.setSmax(Integer.valueOf(value));
+                node.setSmax(Integer.parseInt(value));
             } else if (TTL.equals(name)) {
-                node.setTtl(Integer.valueOf(value));
+                node.setTtl(Integer.parseInt(value));
             } else if (TIMEOUT.equals(name)) {
-                node.setTimeout(Integer.valueOf(value));
+                node.setTimeout(Integer.parseInt(value));
             } else if (CONTEXT.equals(name)) {
                 final String[] context = value.split(",");
                 contexts = Arrays.asList(context);
@@ -375,7 +374,7 @@ class MCMPHandler implements HttpHandler {
             processError(TYPESYNTAX, SMISFLD, exchange);
             return;
         }
-        final List<String> virtualHosts = aliases != null ? Arrays.asList(aliases.split(",")) : null;
+        final List<String> virtualHosts = Arrays.asList(aliases.split(","));
         if (virtualHosts == null || virtualHosts.isEmpty()) {
             processError(TYPESYNTAX, SCONBAD, exchange);
             return;
@@ -440,7 +439,7 @@ class MCMPHandler implements HttpHandler {
         }
 
         UndertowLogger.ROOT_LOGGER.receivedNodeLoad(jvmRoute, loadValue);
-        final int load = Integer.valueOf(loadValue);
+        final int load = Integer.parseInt(loadValue);
         if (load > 0 || load == -2) {
 
             final Node node = container.getNode(jvmRoute);
@@ -555,7 +554,7 @@ class MCMPHandler implements HttpHandler {
                     return;
                 }
                 // Check whether we can reach the host
-                checkHostUp(scheme, host, Integer.valueOf(port), exchange, new NodePingUtil.PingCallback() {
+                checkHostUp(scheme, host, Integer.parseInt(port), exchange, new NodePingUtil.PingCallback() {
                     @Override
                     public void completed() {
                         sendResponse(exchange, OK);
