@@ -489,7 +489,7 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
         try {
             InstanceFactory<T> factory = servletContext.getDeployment().getDeploymentInfo().getClassIntrospecter().createInstanceFactory(handlerClass);
             final InstanceHandle<T> instance = factory.createInstance();
-            exchange.upgradeChannel(new ServletUpgradeListener<>(instance, servletContext.getDeployment().getThreadSetupAction(), exchange));
+            exchange.upgradeChannel(new ServletUpgradeListener<>(instance, servletContext.getDeployment(), exchange));
             return instance.getInstance();
         } catch (InstantiationException e) {
             throw new RuntimeException(e);
@@ -510,7 +510,10 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
                 if(formData != null) {
                     for (final String namedPart : formData) {
                         for (FormData.FormValue part : formData.get(namedPart)) {
-                            parts.add(new PartImpl(namedPart, part, requestContext.getOriginalServletPathMatch().getServletChain().getManagedServlet().getMultipartConfig(), servletContext));
+                            parts.add(new PartImpl(namedPart,
+                                    part,
+                                    requestContext.getOriginalServletPathMatch().getServletChain().getManagedServlet().getMultipartConfig(),
+                                    servletContext, this));
                         }
                     }
                 }
@@ -729,7 +732,6 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
                         }
                     } else {
                         final ArrayList<String> values = new ArrayList<>();
-                        int i = 0;
                         for (final FormData.FormValue v : val) {
                             if(!v.isFile()) {
                                 values.add(v.getValue());
